@@ -1,23 +1,5 @@
 package SQL::Translator::Schema::View;
 
-# ----------------------------------------------------------------------
-# Copyright (C) 2002-2009 SQLFairy Authors
-#
-# This program is free software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License as
-# published by the Free Software Foundation; version 2.
-#
-# This program is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-# 02111-1307  USA
-# -------------------------------------------------------------------
-
 =pod
 
 =head1 NAME
@@ -41,22 +23,15 @@ C<SQL::Translator::Schema::View> is the view object.
 
 =cut
 
-use strict;
-use SQL::Translator::Utils 'parse_list_arg';
+use Moo;
+use SQL::Translator::Utils qw(ex2err);
+use SQL::Translator::Types qw(schema_obj);
+use SQL::Translator::Role::ListAttr;
+use Sub::Quote qw(quote_sub);
 
-use base 'SQL::Translator::Schema::Object';
+extends 'SQL::Translator::Schema::Object';
 
-use vars qw($VERSION $TABLE_COUNT $VIEW_COUNT);
-
-$VERSION = '1.59';
-
-# ----------------------------------------------------------------------
-
-__PACKAGE__->_attributes( qw/
-    name sql fields schema order
-/);
-
-=pod
+our $VERSION = '1.59';
 
 =head2 new
 
@@ -64,13 +39,6 @@ Object constructor.
 
   my $view = SQL::Translator::Schema::View->new;
 
-=cut
-
-# ----------------------------------------------------------------------
-sub fields {
-
-=pod
-
 =head2 fields
 
 Gets and set the fields the constraint is on.  Accepts a string, list or
@@ -87,24 +55,38 @@ names and keep them in order by the first occurrence of a field name.
 
 =cut
 
-    my $self   = shift;
-    my $fields = parse_list_arg( @_ );
+with ListAttr fields => ( uniq => 1 );
 
-    if ( @$fields ) {
-        my ( %unique, @unique );
-        for my $f ( @$fields ) {
-            next if $unique{ $f };
-            $unique{ $f } = 1;
-            push @unique, $f;
-        }
+=head2 tables
 
-        $self->{'fields'} = \@unique;
-    }
+Gets and set the tables the SELECT mentions.  Accepts a string, list or
+arrayref; returns an array or array reference.  Will unique the table
+names and keep them in order by the first occurrence of a field name.
 
-    return wantarray ? @{ $self->{'fields'} || [] } : ($self->{'fields'} || '');
-}
+  $view->tables('foo');
+  $view->tables('foo', 'bar');
+  $view->tables( 'foo, bar' );
+  $view->tables( [ 'foo', 'bar' ] );
+  $view->tables( qw[ foo bar ] );
+
+  my @tables = $view->tables;
+
+=cut
+
+with ListAttr tables => ( uniq => 1 );
+
+=head2 options
+
+Gets and sets a list of options on the view.
+
+  $view->options('ALGORITHM=UNDEFINED');
+
+  my @options = $view->options;
+
+=cut
+
+with ListAttr options => ( uniq => 1, append => 1 );
 
-# ----------------------------------------------------------------------
 sub is_valid {
 
 =pod
@@ -125,11 +107,6 @@ Determine whether the view is valid or not.
     return 1;
 }
 
-# ----------------------------------------------------------------------
-sub name {
-
-=pod
-
 =head2 name
 
 Get or set the view's name.
@@ -138,15 +115,7 @@ Get or set the view's name.
 
 =cut
 
-    my $self        = shift;
-    $self->{'name'} = shift if @_;
-    return $self->{'name'} || '';
-}
-
-# ----------------------------------------------------------------------
-sub order {
-
-=pod
+has name => ( is => 'rw', default => quote_sub(q{ '' }) );
 
 =head2 order
 
@@ -156,19 +125,17 @@ Get or set the view's order.
 
 =cut
 
-    my ( $self, $arg ) = @_;
+has order => ( is => 'rw', default => quote_sub(q{ 0 }) );
+
+around order => sub {
+    my ( $orig, $self, $arg ) = @_;
 
     if ( defined $arg && $arg =~ /^\d+$/ ) {
-        $self->{'order'} = $arg;
+        return $self->$orig($arg);
     }
 
-    return $self->{'order'} || 0;
-}
-
-# ----------------------------------------------------------------------
-sub sql {
-
-=pod
+    return $self->$orig;
+};
 
 =head2 sql
 
@@ -178,15 +145,7 @@ Get or set the view's SQL.
 
 =cut
 
-    my $self       = shift;
-    $self->{'sql'} = shift if @_;
-    return $self->{'sql'} || '';
-}
-
-# ----------------------------------------------------------------------
-sub schema {
-
-=pod
+has sql => ( is => 'rw', default => quote_sub(q{ '' }) );
 
 =head2 schema
 
@@ -197,20 +156,9 @@ Get or set the view's schema object.
 
 =cut
 
-    my $self = shift;
-    if ( my $arg = shift ) {
-        return $self->error('Not a schema object') unless
-            UNIVERSAL::isa( $arg, 'SQL::Translator::Schema' );
-        $self->{'schema'} = $arg;
-    }
+has schema => ( is => 'rw', isa => schema_obj('Schema'), weak_ref => 1 );
 
-    return $self->{'schema'};
-}
-
-# ----------------------------------------------------------------------
-sub equals {
-
-=pod
+around schema => \&ex2err;
 
 =head2 equals
 
@@ -220,15 +168,17 @@ Determines if this view is the same as another
 
 =cut
 
+around equals => sub {
+    my $orig = shift;
     my $self = shift;
     my $other = shift;
     my $case_insensitive = shift;
     my $ignore_sql = shift;
-    
-    return 0 unless $self->SUPER::equals($other);
+
+    return 0 unless $self->$orig($other);
     return 0 unless $case_insensitive ? uc($self->name) eq uc($other->name) : $self->name eq $other->name;
     #return 0 unless $self->is_valid eq $other->is_valid;
-    
+
     unless ($ignore_sql) {
         my $selfSql = $self->sql;
         my $otherSql = $other->sql;
@@ -240,24 +190,19 @@ Determines if this view is the same as another
         $otherSql =~ s/\s+/ /sg;
         return 0 unless $selfSql eq $otherSql;
     }
-    
+
     my $selfFields = join(":", $self->fields);
     my $otherFields = join(":", $other->fields);
     return 0 unless $case_insensitive ? uc($selfFields) eq uc($otherFields) : $selfFields eq $otherFields;
     return 0 unless $self->_compare_objects(scalar $self->extra, scalar $other->extra);
     return 1;
-}
+};
 
-# ----------------------------------------------------------------------
-sub DESTROY {
-    my $self = shift;
-    undef $self->{'schema'}; # destroy cyclical reference
-}
+# Must come after all 'has' declarations
+around new => \&ex2err;
 
 1;
 
-# ----------------------------------------------------------------------
-
 =pod
 
 =head1 AUTHOR
